#![allow(clippy::type_complexity)]

use super::engine_event_handling::*;
use super::engine_settings::EngineSettings;

use anyhow::Result;
use assetpath::AssetPath;
use context::prelude::cgmath::vec3;
use ecs::{Resource, resources::Resource as ResourceTrait};

use crate::assets::asset_manager::AssetManager;
use crate::prelude::*;

use std::collections::HashMap;
use std::sync::Arc;

#[derive(Resource)]
pub struct GuiHandlerRenderer;

impl TScene for GuiHandlerRenderer {
    fn process(
        &mut self,
        buffer_recorder: &mut CommandBufferRecorder<'_>,
        _images: &TargetMode<Vec<Arc<Image>>>,
        indices: &TargetMode<usize>,
        world: &mut World,
    ) -> Result<()> {
        let gui_handler: &mut GuiHandler = world.resources.get_mut()?;
        gui_handler.process(buffer_recorder, indices)
    }

    fn resize(
        &mut self,
        window_width: f32,
        window_height: f32,
        images: &TargetMode<Vec<Arc<Image>>>,
        world: &mut World,
    ) -> Result<()> {
        let gui_handler: &mut GuiHandler = world.resources.get_mut()?;
        gui_handler.resize(window_width as u32, window_height as u32, images)
    }
}

#[derive(Resource)]
pub struct Engine {
    pub(crate) resource_base_path: String,
}

impl Engine {
    pub fn new<T: ResourceTrait + EventConsumer>(
        #[allow(unused)] mut create_info: EngineCreateInfo<'_>,
        world: &mut WorldBuilder,
    ) -> Result<()> {
        create_info
            .controller_directories
            .xbox_path
            .as_mut()
            .map(|p| p.set_prefix(&create_info.resource_base_path));
        create_info
            .controller_directories
            .steam_path
            .as_mut()
            .map(|p| p.set_prefix(&create_info.resource_base_path));
        create_info
            .controller_directories
            .ps4_path
            .as_mut()
            .map(|p| p.set_prefix(&create_info.resource_base_path));

        create_info
            .asset_directories
            .entity_file_directory
            .set_prefix(&create_info.resource_base_path);
        create_info
            .asset_directories
            .gltf_file_directory
            .set_prefix(&create_info.resource_base_path);
        create_info
            .asset_directories
            .tile_file_directory
            .set_prefix(&create_info.resource_base_path);

        create_info
            .key_backgrounds
            .light
            .set_prefix(&create_info.resource_base_path);
        create_info
            .key_backgrounds
            .dark
            .set_prefix(&create_info.resource_base_path);

        // ---------------  context  --------------------

        let mut context = {
            create_info.app_info.engine_name = "EngineNameNeeded".to_string();
            create_info.app_info.engine_version = 1;

            // gui handler needs color attachment
            // rest is depending on render type
            let image_usage = VK_IMAGE_USAGE_TRANSFER_DST_BIT | VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT;

            let device_features = DeviceFeatures {
                samplerAnisotropy: VK_TRUE,

                ..Default::default()
            };

            let mut context_builder = Context::new()
                .set_volume_info(create_info.volume_info)
                .set_app_info(create_info.app_info)
                .set_window_info(create_info.window_info)
                .set_os_specific_info(create_info.os_specific_config)
                .set_vulkan_debug_info(create_info.vulkan_debug_info)
                .set_controller_deadzones(create_info.controller_deadzones)
                .set_sample_count(create_info.graphics_info.sample_count)
                .set_render_core_info(
                    VK_FORMAT_R8G8B8A8_SRGB,
                    image_usage,
                    create_info.graphics_info.vsync,
                )
                .set_device_features(device_features);

            if create_info.enable_mouse {
                context_builder = context_builder.enable_mouse();
            }

            if create_info.enable_keyboard {
                context_builder = context_builder.enable_keyboard();
            }

            if create_info.enable_controller {
                context_builder = context_builder.enable_controller();
            }

            if let SceneType::Raytracer = create_info.graphics_info.render_type {
                context_builder = context_builder.enable_ray_tracing();
            }

            if create_info.enable_backtrace {
                context_builder = context_builder.enable_backtrace();
            }

            context_builder.build::<Scene>()?
        };

        // update sound handler center
        context
            .sound()
            .set_direction(vec3(0.0, 1.0, 0.0), vec3(0.0, 0.0, 1.0))?;

        // ---------------  settings  -------------------

        {
            create_info.graphics_info.render_type = Scene::check_scene_type(
                context.device(),
                create_info.graphics_info.render_type,
                create_info.rasterizer_info.use_deferred,
            )
            .ok_or_else(|| anyhow::Error::msg("No proper render type could be found"))?;
        }

        let engine_settings = EngineSettings::new(
            &context,
            create_info.graphics_info,
            create_info.raytracing_info,
            create_info.rasterizer_info,
            create_info.controller_directories,
            create_info.asset_directories,
            create_info.key_backgrounds,
            create_info.resource_base_path.clone(),
        )?;

        let asset_manager = AssetManager::new(&engine_settings)?;
        let gui_handler = GuiHandler::new(create_info.gui_info, &context)?;

        context
            .render_core_mut()
            .add_render_routine::<GuiHandlerRenderer>(10_000_000);
        context.render_core_mut().add_render_routine::<Scene>(100);

        world.resources.insert(context);

        Scene::new(
            create_info.rasterizer_info,
            create_info.raytracing_info,
            create_info.graphics_info,
            world,
        )?;

        // default keyboard navigation
        let mut direction_mapping = HashMap::new();
        direction_mapping.insert(Keycode::Left, GuiDirection::Left);
        direction_mapping.insert(Keycode::Right, GuiDirection::Right);
        direction_mapping.insert(Keycode::Up, GuiDirection::Up);
        direction_mapping.insert(Keycode::Down, GuiDirection::Down);

        world.resources.insert(gui_handler);
        world.resources.insert(GuiHandlerRenderer);
        world.resources.insert(InputMap { direction_mapping });
        world.resources.insert(asset_manager);
        world.resources.insert(Engine {
            resource_base_path: create_info.resource_base_path,
        });
        world.resources.insert(engine_settings);

        world.add_system(1_000_000, Self::main_system::<T>);

        Ok(())
    }

    // pub fn set_game_object<E>(&self, engine_object: Option<E>) -> Result<()>
    // where
    //     E: EngineObject + Send + Sync,
    // {
    //     self.context().set_context_object(match engine_object {
    //         Some(engine_object) => Some(ContextObjectImpl::new(
    //             self.graphical.context.clone(),
    //             engine_object,
    //             self.graphical.input.clone(),
    //             self.gui_handler().clone(),
    //         )?),
    //         None => None,
    //     });

    //     Ok(())
    // }

    // pub fn loading_screen<T, R, L, G>(
    //     &self,
    //     gui: Option<Arc<G>>,
    //     loader: L,
    //     on_ready: R,
    // ) -> Result<()>
    // where
    //     R: Fn(T) -> anyhow::Result<()> + 'static + Send + Sync,
    //     T: Send + Sync + 'static,
    //     L: FnOnce() -> T + Send + Sync + 'static,
    //     G: TopLevelGui + TopGui + Send + Sync + 'static,
    // {
    //     LoadingScreen::load(self.context(), gui, loader, on_ready)
    // }

    pub fn build_path(&self, path: &str) -> AssetPath {
        (self.resource_base_path.as_str(), path).into()
    }
}

impl Engine {
    fn main_system<T: ResourceTrait + EventConsumer>(world: &mut World) -> Result<bool> {
        let gui_handler = world.resources.get_mut_unchecked::<GuiHandler>();
        let input_map = world.resources.get_unchecked::<InputMap>();
        let context = world.resources.get_unchecked::<Context>();

        let res = world.resources.get_mut_unchecked::<Context>().next_frame(
            world,
            |world, consumer: &mut T, event| {
                Self::event(world, context, gui_handler, input_map, consumer, event)
            },
        )?;

        gui_handler.process_callbacks(world)?;

        Ok(res)
    }
}
