use crate::prelude::*;

use asset::GltfBoundingBox as GltfBB;

use serde::{Deserialize, Serialize};
use utilities::prelude::cgmath::Vector3;

#[derive(Debug, Default, Clone, PartialEq, Serialize, Deserialize)]
pub struct BoundingBox {
    pub min: [f32; 3],
    pub max: [f32; 3],
}

impl BoundingBox {
    pub fn init() -> Self {
        Self {
            min: [f32::INFINITY; 3],
            max: [-f32::INFINITY; 3],
        }
    }

    pub fn check(&mut self, v: Vector3<f32>) {
        self.min[0] = self.min[0].min(v.x);
        self.min[1] = self.min[1].min(v.y);
        self.min[2] = self.min[2].min(v.z);

        self.max[0] = self.max[0].max(v.x);
        self.max[1] = self.max[1].max(v.y);
        self.max[2] = self.max[2].max(v.z);
    }

    pub fn is_zero(&self) -> bool {
        self.min[0] == 0.0
            && self.min[1] == 0.0
            && self.min[2] == 0.0
            && self.max[0] == 0.0
            && self.max[1] == 0.0
            && self.max[2] == 0.0
    }

    pub fn check_dimensions(&mut self, min_extent: f32) {
        for i in 0..3 {
            let distance = self.min[i].abs() + self.max[i].abs();

            if distance < min_extent {
                self.min[i] = self.min[i].min(-(min_extent / 2.0));
                self.max[i] = self.max[i].max(min_extent / 2.0);
            }
        }
    }

    pub fn corners(&self) -> [Vector3<f32>; 8] {
        let min_x = self.min[0];
        let min_y = self.min[1];
        let min_z = self.min[2];

        let max_x = self.max[0];
        let max_y = self.max[1];
        let max_z = self.max[2];

        [
            // front bottom left
            Vector3::new(min_x, min_y, min_z),
            // front bottom right
            Vector3::new(max_x, min_y, min_z),
            // front top left
            Vector3::new(min_x, min_y, max_z),
            // front top right
            Vector3::new(max_x, min_y, max_z),
            // back bottom left
            Vector3::new(min_x, max_y, min_z),
            // back bottom right
            Vector3::new(max_x, max_y, min_z),
            // back top left
            Vector3::new(min_x, max_y, max_z),
            // back top right
            Vector3::new(max_x, max_y, max_z),
        ]
    }

    pub fn to_box(&self) -> [Vector3<f32>; 36] {
        let corners = self.corners();

        [
            // front
            corners[1], corners[3], corners[0], // 0
            corners[3], corners[2], corners[0], // 1
            // bottom
            corners[1], corners[5], corners[0], // 2
            corners[5], corners[4], corners[0], // 3
            // right
            corners[1], corners[5], corners[3], // 4
            corners[5], corners[7], corners[3], // 5
            // left
            corners[0], corners[2], corners[4], // 6
            corners[4], corners[2], corners[6], // 7
            // top
            corners[3], corners[7], corners[2], // 8
            corners[7], corners[6], corners[2], // 9
            // back
            corners[5], corners[4], corners[7], // 10
            corners[4], corners[6], corners[7], // 11
        ]
    }

    pub fn min_vec(&self) -> Vector3<f32> {
        Vector3::from(self.min)
    }

    pub fn max_vec(&self) -> Vector3<f32> {
        Vector3::from(self.max)
    }
}

impl From<GltfBB> for BoundingBox {
    fn from(bb: GltfBB) -> Self {
        Self {
            min: bb.min,
            max: bb.max,
        }
    }
}

impl EntityComponent for BoundingBox {
    fn name(&self) -> &str {
        Self::debug_name()
    }
}

impl ComponentDebug for BoundingBox {
    fn debug_name() -> &'static str {
        "BoundingBox"
    }
}

impl ComponentCreateInfo<'_> for BoundingBox {}
