mod deferredhandler;

use crate::{prelude::*, update_iter};

use anyhow::Result;
use asset::RasterizerDescriptorLayouts as AssetLayouts;
use paste;
use utilities::prelude::cgmath::{Vector3, Zero};

use super::{
    super::{
        ExtensionCheck, RasterizerDescriptorLayouts, RenderingFrontEnd,
        shared::{bufferhandler::*, safecommandbuffer::SafeCommandBuffer},
    },
    SceneInfo,
    lights::light_wrapper::LightHandler,
    pipelines::RasterizerPipelines,
    rasterizershader::RasterizerShader,
};

use deferredhandler::{DeferredHandler, FallbackInfo};

use std::sync::{Arc, Mutex};

pub struct Rasterizer {
    device: Arc<Device>,
    queue: Arc<Mutex<Queue>>,

    user_view: View,

    light_handler: LightHandler,

    // scene info
    scene_info_buffer: Arc<Buffer<SceneInfo>>,
    deferred_descriptor: Arc<DescriptorSet>,

    deferred_handler: DeferredHandler,

    buffer_handler: BufferHandler,
    multi_render_target: RenderTarget,

    fallback: bool,

    enable_lighting: bool,
    render_scale: f32,
}

impl Rasterizer {
    pub fn new(
        device: &Arc<Device>,
        queue: &Arc<Mutex<Queue>>,
        (screen_width, screen_height): (f32, f32),
        images: &TargetMode<Vec<Arc<Image>>>,
        render_scale: f32,
        rasterizer_info: RasterizerInfo,
        enforce_fallback: bool,
    ) -> Result<Self> {
        let proj = perspective(45.0, screen_width / screen_height, 0.01, 100.0);
        let view = View::new(device, proj, 45.0, true)?;

        let (multi_render_target, buffer_handler, deferred_handler) = Self::create_target_handlers(
            device,
            queue,
            (screen_width, screen_height),
            images,
            render_scale,
            if enforce_fallback {
                Some(FallbackInfo::default())
            } else {
                None
            },
        )?;

        let scene_info = SceneInfo {
            point_light_count: 0,
            directional_light_count: 0,
            spot_light_count: 0,
            view_position: Vector3::zero(),
        };

        let scene_info_buffer = Buffer::builder()
            .set_usage(VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT)
            .set_memory_usage(MemoryUsage::CpuOnly)
            .set_data(&[scene_info])
            .build(device.clone())?;

        let deferred_descriptor = Self::create_deferred_descriptor(device, &scene_info_buffer)?;
        let attachments = multi_render_target.sub_pass(0).attachments();

        let world_pos_map = attachments[0].image(0);
        let world_normal = attachments[1].image(0);
        let albedo = attachments[2].image(0);
        Self::update_descriptor_images(&deferred_descriptor, world_pos_map, world_normal, albedo)?;

        Ok(Rasterizer {
            user_view: view,

            light_handler: LightHandler::new(
                device,
                queue,
                images.mono().len(),
                enforce_fallback,
                rasterizer_info.shadow_image_size,
            )?,

            scene_info_buffer,

            deferred_descriptor,

            deferred_handler,
            buffer_handler,

            multi_render_target,

            fallback: enforce_fallback,

            device: device.clone(),
            queue: queue.clone(),

            enable_lighting: rasterizer_info.enable_lighting,
            render_scale,
        })
    }

    fn invalidate(&mut self) {
        self.buffer_handler.invalidate();
        self.deferred_handler.invalidate();

        self.light_handler.invalidate();
    }

    fn create_deferred_descriptor(
        device: &Arc<Device>,
        scene_info: &Arc<Buffer<SceneInfo>>,
    ) -> Result<Arc<DescriptorSet>> {
        let descriptor_layout =
            RasterizerDescriptorLayouts::deferred_descriptor_layout(device.clone())?;
        let descriptor_pool = DescriptorPool::builder()
            .set_layout(descriptor_layout)
            .build(device.clone())?;

        let descriptor_set = DescriptorPool::prepare_set(&descriptor_pool).allocate()?;

        descriptor_set.update(&[DescriptorWrite::uniform_buffers(3, &[scene_info])])?;

        Ok(descriptor_set)
    }

    fn update_descriptor_images(
        set: &Arc<DescriptorSet>,
        world_position: &Arc<Image>,
        world_normal: &Arc<Image>,
        albedo: &Arc<Image>,
    ) -> Result<()> {
        set.update(&[
            DescriptorWrite::combined_samplers(0, &[world_position])
                .change_image_layout(VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL),
            DescriptorWrite::combined_samplers(1, &[world_normal])
                .change_image_layout(VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL),
            DescriptorWrite::combined_samplers(2, &[albedo])
                .change_image_layout(VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL),
        ])?;

        Ok(())
    }

    fn create_target_handlers(
        device: &Arc<Device>,
        queue: &Arc<Mutex<Queue>>,
        (screen_width, screen_height): (f32, f32),
        images: &TargetMode<Vec<Arc<Image>>>,
        render_scale: f32,
        fallback: Option<FallbackInfo>,
    ) -> Result<(RenderTarget, BufferHandler, DeferredHandler)> {
        let width = (screen_width * render_scale) as u32;
        let height = (screen_height * render_scale) as u32;

        let images = images.mono();

        let multi_render_target = RenderTarget::builder()
            .add_sub_pass(
                SubPass::builder(width, height)
                    // (World space) Positions
                    .add_target_info(CustomTarget {
                        usage: VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT.into(),
                        format: VK_FORMAT_R16G16B16A16_SFLOAT,
                        clear_on_load: true,
                        store_on_save: true,
                        attach_sampler: true,
                        use_as_input: false,
                        clear_value: ClearValue::Color([0.0, 0.0, 0.0, 0.0]),
                    })
                    // (World space) Normals
                    .add_target_info(CustomTarget {
                        usage: VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT.into(),
                        format: VK_FORMAT_R16G16B16A16_SFLOAT,
                        clear_on_load: true,
                        store_on_save: true,
                        attach_sampler: true,
                        use_as_input: false,
                        clear_value: ClearValue::Color([0.0, 0.0, 0.0, 0.0]),
                    })
                    // Albedo (color)
                    .add_target_info(CustomTarget {
                        usage: VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT.into(),
                        format: VK_FORMAT_R8G8B8A8_UNORM,
                        clear_on_load: true,
                        store_on_save: true,
                        attach_sampler: true,
                        use_as_input: false,
                        clear_value: ClearValue::Color([0.0, 0.0, 0.0, 0.0]),
                    })
                    // Depth attachment
                    .add_target_info(CustomTarget::depth())
                    .set_output_usage(VK_ACCESS_SHADER_READ_BIT)
                    .use_queue(queue.clone())
                    .build(device)?,
            )
            .add_sub_pass(
                SubPass::builder(width, height)
                    .set_prepared_targets(images, 0, None)
                    .build(device)?,
            )
            .build(device)?;

        // create all descriptor set layouts, that are needed for the pipeline layout
        let globals_layout = View::desc_layout(device)?;
        let model_layout = AssetLayouts::descriptor_layout(device.clone())?;
        let model_bone_layout = AssetLayouts::bone_descriptor_layout(device.clone())?;
        let model_color_layout = AssetLayouts::color_descriptor_layout(device.clone())?;
        let model_color_bone_layout = AssetLayouts::color_bone_descriptor_layout(device.clone())?;
        let model_color_normal_layout =
            AssetLayouts::color_normal_descriptor_layout(device.clone())?;
        let model_color_normal_bone_layout =
            AssetLayouts::color_bone_normal_descriptor_layout(device.clone())?;

        let shader = RasterizerShader::mrt(device)?;
        let attachment_info = vec![false, false, false];

        let create_info = BufferHandlerCreateInfo::new(
            &shader,
            device,
            queue,
            images.len(),
            BufferHandlerType::graphics(
                multi_render_target.render_pass(),
                &attachment_info,
                0,
                RasterizerPipelines::create_default_pipeline::<PositionOnly>,
                PipelineLayoutInfo {
                    descriptor_layouts: vec![model_layout.as_ref(), globals_layout.as_ref()],
                    push_constants: &[],
                },
                RasterizerPipelines::create_default_pipeline::<PositionColorUV>,
                PipelineLayoutInfo {
                    descriptor_layouts: vec![model_color_layout.as_ref(), globals_layout.as_ref()],
                    push_constants: &[],
                },
                RasterizerPipelines::create_default_pipeline::<PositionNormal>,
                PipelineLayoutInfo {
                    descriptor_layouts: vec![model_layout.as_ref(), globals_layout.as_ref()],
                    push_constants: &[],
                },
                RasterizerPipelines::create_default_pipeline::<PositionNormalBone>,
                PipelineLayoutInfo {
                    descriptor_layouts: vec![model_bone_layout.as_ref(), globals_layout.as_ref()],
                    push_constants: &[],
                },
                RasterizerPipelines::create_default_pipeline::<PositionNormalColor>,
                PipelineLayoutInfo {
                    descriptor_layouts: vec![model_layout.as_ref(), globals_layout.as_ref()],
                    push_constants: &[],
                },
                RasterizerPipelines::create_default_pipeline::<PositionNormalColorUV>,
                PipelineLayoutInfo {
                    descriptor_layouts: vec![model_color_layout.as_ref(), globals_layout.as_ref()],
                    push_constants: &[],
                },
                RasterizerPipelines::create_default_pipeline::<PositionNormalColorUVBone>,
                PipelineLayoutInfo {
                    descriptor_layouts: vec![
                        model_color_bone_layout.as_ref(),
                        globals_layout.as_ref(),
                    ],
                    push_constants: &[],
                },
                RasterizerPipelines::create_default_pipeline::<PositionNormalColorUVNormalUV>,
                PipelineLayoutInfo {
                    descriptor_layouts: vec![
                        model_color_normal_layout.as_ref(),
                        globals_layout.as_ref(),
                    ],
                    push_constants: &[],
                },
                RasterizerPipelines::create_default_pipeline::<PositionNormalColorUVNormalUVBone>,
                PipelineLayoutInfo {
                    descriptor_layouts: vec![
                        model_color_normal_bone_layout.as_ref(),
                        globals_layout.as_ref(),
                    ],
                    push_constants: &[],
                },
            ),
        );

        let mut buffer_handler = BufferHandler::new(&create_info, true)?;
        let deferred_handler = DeferredHandler::new(
            device,
            queue,
            images.len(),
            multi_render_target.render_pass(),
            fallback,
        )?;

        update_iter!(
            buffer_handler,
            (position, PositionOnly),
            (position_coloruv, PositionColorUV),
            (position_normal, PositionNormal),
            (position_normal_coloruv, PositionNormalColorUV),
            (position_normal_coloruv_bone, PositionNormalColorUVBone),
            (
                position_normal_coloruv_normaluv,
                PositionNormalColorUVNormalUV
            ),
            (
                position_normal_coloruv_normaluv_bone,
                PositionNormalColorUVNormalUVBone
            ),
        );

        Ok((multi_render_target, buffer_handler, deferred_handler))
    }

    fn recompile_fallback_shaders(&mut self) -> Result<()> {
        if self.fallback {
            let (point, directional, spot) = self.light_handler.light_data();

            self.deferred_handler.resize_fallback_shader(
                FallbackInfo {
                    spot_light_count: spot.light_count,
                    directional_light_count: directional.light_count,
                    point_light_count: point.light_count,
                },
                self.multi_render_target.render_pass(),
            )?;
        }

        Ok(())
    }
}

impl ExtensionCheck for Rasterizer {
    fn check(device: &Arc<Device>) -> bool {
        let physical_device = device.physical_device();

        if physical_device
            .descriptor_indexing_features()
            .descriptorBindingVariableDescriptorCount
            == VK_FALSE
        {
            return false;
        }

        let properties = device.physical_device().properties();

        if properties.limits.maxBoundDescriptorSets < 7 {
            return false;
        }

        true
    }
}

impl RenderingFrontEnd for Rasterizer {
    fn render_scale(&self) -> f32 {
        self.render_scale
    }

    // user point of view handle
    fn view_mut(&mut self) -> &mut View {
        &mut self.user_view
    }

    fn view(&self) -> &View {
        &self.user_view
    }

    fn screen_to_world(&self, _x: u32, _y: u32) -> Result<Option<Vector3<f32>>> {
        todo!()
    }

    fn set_center(&mut self, center: cgmath::Vector3<f32>) -> Result<()> {
        if self.user_view.camera_mut().set_center(center) {
            self.user_view.update_buffer()?;
        }

        self.light_handler.set_center(center)?;

        // update user view
        let mut mapped = self.scene_info_buffer.map_complete()?;
        mapped[0].view_position = center;

        Ok(())
    }

    fn process(
        &mut self,
        _changed: bool,
        content: Vec<&EntityObject>,
        buffer_recorder: &mut CommandBufferRecorder<'_>,
        _images: &TargetMode<Vec<Arc<Image>>>,
        indices: &TargetMode<usize>,
    ) -> Result<()> {
        // better handling (maýbe not always a rerecord is required)
        self.invalidate();

        let index = *indices.mono();

        self.light_handler
            .render(&content, buffer_recorder, index)?;

        // render multi render target
        self.multi_render_target.begin(
            buffer_recorder,
            VK_SUBPASS_CONTENTS_SECONDARY_COMMAND_BUFFERS,
            index,
        );

        // let mut list = Vec::new();

        // self.buffer_handler.execute(
        //     Some(self.multi_render_target.framebuffer(index)),
        //     content,
        //     &[],
        //     &mut list,
        //     index,
        //     &self.user_view,
        // )?;

        println!("deferred renderer is not implemented!");

        // buffer_recorder.execute_commands(&list);

        self.multi_render_target.next_subpass(
            buffer_recorder,
            VK_SUBPASS_CONTENTS_SECONDARY_COMMAND_BUFFERS,
        );

        let (point, directional, spot) = self.light_handler.light_data();

        self.deferred_handler.render(
            buffer_recorder,
            &[
                &self.deferred_descriptor,
                point.info_descriptor,
                point.depth_map_descriptor,
                directional.info_descriptor,
                directional.depth_map_descriptor,
                spot.info_descriptor,
                spot.depth_map_descriptor,
            ],
            self.multi_render_target.framebuffer(index),
            index,
        )?;

        self.multi_render_target.end(buffer_recorder);

        Ok(())
    }

    fn resize(
        &mut self,
        window_width: f32,
        window_height: f32,
        images: &TargetMode<Vec<Arc<Image>>>,
    ) -> Result<()> {
        let fov = 45.0;
        let proj = perspective(fov, window_width / window_height, 0.01, 100.0);
        self.user_view = View::new_proj(&self.user_view, proj, fov)?;

        let (point, directional, spot) = self.light_handler.light_data();

        let (multi_render_target, buffer_handler, deferred_handler) = Self::create_target_handlers(
            &self.device,
            &self.queue,
            (window_width, window_height),
            images,
            self.render_scale,
            if self.fallback {
                Some(FallbackInfo {
                    spot_light_count: spot.light_count,
                    directional_light_count: directional.light_count,
                    point_light_count: point.light_count,
                })
            } else {
                None
            },
        )?;

        let attachments = multi_render_target.sub_pass(0).attachments();

        let world_pos_map = attachments[0].image(0);
        let world_normal = attachments[1].image(0);
        let albedo = attachments[2].image(0);
        Self::update_descriptor_images(
            &self.deferred_descriptor,
            world_pos_map,
            world_normal,
            albedo,
        )?;

        self.multi_render_target = multi_render_target;
        self.buffer_handler = buffer_handler;
        self.deferred_handler = deferred_handler;

        Ok(())
    }

    // light handling
    fn add_light(&mut self, light: Light) -> Result<()> {
        if self.enable_lighting {
            self.light_handler.add(light)?;

            {
                let (point, directional, spot) = self.light_handler.light_data();

                let mut mapped = self.scene_info_buffer.map_complete()?;
                mapped[0].directional_light_count = directional.light_count;
                mapped[0].point_light_count = spot.light_count;
                mapped[0].spot_light_count = point.light_count;
            }

            self.recompile_fallback_shaders()?;
        }

        Ok(())
    }

    fn remove_light(&mut self, light: Light) -> Result<()> {
        if self.enable_lighting && self.light_handler.remove(&light)? {
            {
                let (point, directional, spot) = self.light_handler.light_data();

                let mut mapped = self.scene_info_buffer.map_complete()?;
                mapped[0].directional_light_count = directional.light_count;
                mapped[0].point_light_count = spot.light_count;
                mapped[0].spot_light_count = point.light_count;
            }

            self.recompile_fallback_shaders()?;
        }

        Ok(())
    }

    fn clear_lights(&mut self) -> Result<()> {
        if self.enable_lighting && self.light_handler.clear()? {
            {
                let (point, directional, spot) = self.light_handler.light_data();

                let mut mapped = self.scene_info_buffer.map_complete()?;
                mapped[0].directional_light_count = directional.light_count;
                mapped[0].point_light_count = spot.light_count;
                mapped[0].spot_light_count = point.light_count;
            }

            self.recompile_fallback_shaders()?;
        }

        Ok(())
    }
}
