#version 460
#extension GL_EXT_ray_tracing : enable
#extension GL_GOOGLE_include_directive : enable

#ifndef RECURSION_DEPTH
    #define RECURSION_DEPTH 2
#endif

#define AMBIENT 0.2

#include "../global_data.glsl"

struct RayPayload {
	vec4 color;
	float distance;
    int depth;
};

struct ShadowPayload {
    float alpha;
};

layout(set = 0, binding = 0) uniform accelerationStructureEXT tlas;

layout (location = 0) rayPayloadInEXT RayPayload pay_load;
layout (location = 1) rayPayloadEXT ShadowPayload shadow_pay_load;
hitAttributeEXT vec2 attribs;

void own_implementation() {
    Triangle triangle = current_triangle(gl_PrimitiveID, attribs);

    // get color
    vec4 tmp = color(triangle);

    vec3 diffuse = tmp.xyz;
    float alpha = tmp.w;
    float max_alpha = alpha;

    // get normal
    vec3 normal = normal(triangle);

    // get roughness / metalness
    float metallic_factor, roughness_factor;

    metallic_roughness(triangle, metallic_factor, roughness_factor);

    // get emission color
    vec3 emissive_factor = emissive_factor(triangle);

    // shadow casting
    vec3 light_factor = vec3(0.0);

    vec3 origin = gl_WorldRayOriginEXT + gl_WorldRayDirectionEXT * gl_HitTEXT;

    {
        // set trace parameter
        uint ray_flags = gl_RayFlagsSkipClosestHitShaderEXT;
        uint cull_mask = 0xff;
        const int pay_load_index = 1;
        int hit_shader_index = 0;
        int miss_shader_index = 1;
        int sbt_record_stride = 1;
        float tmin = 0.0001;

        for (uint i = 0; i < light_info.point_light_count; i++) {
            vec3 direction = lights[nonuniformEXT(i)].position.xyz - origin;

            float dir_length = length(direction);
            direction = direction / dir_length;
            float intensity = intensity(lights[nonuniformEXT(i)].power, dir_length);

            float n_dot_l = dot(direction, normal);

            float factor = 0.0;

            if (intensity > 0.03 && n_dot_l > 0.0) {
                shadow_pay_load.alpha = 0.0;

                traceRayEXT(
                    tlas,
                    ray_flags,
                    cull_mask,
                    hit_shader_index,
                    sbt_record_stride,
                    miss_shader_index,
                    origin,
                    tmin + lights[nonuniformEXT(i)].t_min,
                    direction,
                    dir_length,
                    pay_load_index
                );

                factor = (1.0 - shadow_pay_load.alpha) * n_dot_l * intensity;
            }

            light_factor = max(light_factor, lights[nonuniformEXT(i)].color * factor);
        }
    }

    vec3 object_color = diffuse * rescale(light_factor, 0.0, 1.0, AMBIENT, 1.0);

    float reflector = metallic_factor * (1.0 - roughness_factor);

    // reflection
    if (reflector > 0.0 && pay_load.depth < RECURSION_DEPTH)
    {
        uint ray_flags = gl_RayFlagsOpaqueEXT;
        uint cull_mask = 0xff;
        const int pay_load_index = 0;
        int hit_shader_index = 0;
        int miss_shader_index = 0;
        int sbt_record_stride = 0;
        float tmin = 0.001;
        float tmax = 1000000000000.0;

        vec3 direction = reflect(gl_WorldRayDirectionEXT, normal);

        pay_load.depth++;

        traceRayEXT(
            tlas,
            ray_flags,
            cull_mask,
            hit_shader_index,
            sbt_record_stride,
            miss_shader_index,
            origin,
            tmin,
            direction,
            tmax,
            pay_load_index
        );

        pay_load.depth--;

        object_color = mix(object_color, pay_load.color.xyz, reflector);
    }

    // transparency
    if (alpha < 1.0 && pay_load.depth < RECURSION_DEPTH) {
        uint ray_flags = gl_RayFlagsOpaqueEXT;
        uint cull_mask = 0xff;
        const int pay_load_index = 0;
        int hit_shader_index = 0;
        int miss_shader_index = 0;
        int sbt_record_stride = 0;
        float tmin = 0.001;
        float tmax = 1000000000000.0;

        pay_load.depth++;

        traceRayEXT(
            tlas,
            ray_flags,
            cull_mask,
            hit_shader_index,
            sbt_record_stride,
            miss_shader_index,
            origin,
            tmin,
            gl_WorldRayDirectionEXT,
            tmax,
            pay_load_index
        );

        pay_load.depth--;

        if (pay_load.distance != -1.0) {
            object_color = mix(pay_load.color.xyz, object_color, alpha);
            max_alpha = max(max_alpha, pay_load.color.w);
        }
    }

    // color is at least emissive factor color
    vec3 resulting_color = max(object_color, diffuse * emissive_factor);

    pay_load.color = vec4(resulting_color, max_alpha);
    pay_load.distance = gl_HitTEXT;
}

void pbr_implementation() {
    Triangle triangle = current_triangle(gl_PrimitiveID, attribs);

    vec3 world_position = gl_WorldRayOriginEXT + gl_WorldRayDirectionEXT * gl_HitTEXT;

    PBRInfo pbr_info = prepare(triangle, gl_WorldRayOriginEXT, world_position);

    vec4 color = vec4(0.0);

    {
        // set trace parameter
        uint ray_flags = gl_RayFlagsSkipClosestHitShaderEXT;
        uint cull_mask = 0xff;
        const int pay_load_index = 1;
        int hit_shader_index = 0;
        int miss_shader_index = 1;
        int sbt_record_stride = 1;
        float tmin = 0.0001;

        for (uint i = 0; i < light_info.point_light_count; i++) {
            vec3 direction = lights[nonuniformEXT(i)].position.xyz - world_position;

            float dir_length = length(direction);
            direction = direction / dir_length;
            float intensity = intensity(lights[nonuniformEXT(i)].power, dir_length);

            if (intensity > 0.03) {
                shadow_pay_load.alpha = 0.0;

                traceRayEXT(
                    tlas,
                    ray_flags,
                    cull_mask,
                    hit_shader_index,
                    sbt_record_stride,
                    miss_shader_index,
                    world_position,
                    tmin,
                    direction,
                    dir_length,
                    pay_load_index
                );

                color += apply_light(
                    pbr_info,
                    lights[nonuniformEXT(i)].position.xyz,
                    lights[nonuniformEXT(i)].color,
                    intensity
                );
            }
        }
    }

    pay_load.color = color;
    pay_load.distance = gl_HitTEXT;
}

void main()
{
    own_implementation();
}
