// needed since RLS won't accept #[repr(C)]
#![allow(improper_ctypes)]

use anyhow::{Error, anyhow};

use sdl3::{
    self, Sdl,
    clipboard::ClipboardUtil,
    messagebox::{MessageBoxFlag, show_simple_message_box},
    mouse::Cursor,
    surface::Surface as SDL_Surface,
    video::{FullscreenType, WindowPos},
};

use crate::Result;
use vulkan_rs::prelude::*;

use std::ops::Deref;
use std::path::Path;
use std::sync::{
    Arc, Mutex,
    atomic::{AtomicI32, AtomicU32, Ordering::SeqCst},
};

#[derive(Default, Debug, Clone)]
pub struct WindowCreateInfo {
    pub title: String,
    pub width: u32,
    pub height: u32,
    pub fullscreen: bool,
}

#[derive(Debug)]
pub struct Display {
    pub name: String,

    // bounds
    pub x: i32,
    pub y: i32,
    pub w: u32,
    pub h: u32,
}

#[derive(Default, Debug)]
struct CellRect {
    x: AtomicI32,
    y: AtomicI32,
    width: AtomicU32,
    height: AtomicU32,
}

impl CellRect {
    fn update_from_window(&self, window: &sdl3::video::Window) {
        let (w, h) = window.size();
        let (x, y) = window.position();

        self.x.store(x, SeqCst);
        self.y.store(y, SeqCst);
        self.width.store(w, SeqCst);
        self.height.store(h, SeqCst);
    }

    fn update_to_window(&self, window: &mut sdl3::video::Window) -> Result<()> {
        set_window_size(window, self.width.load(SeqCst), self.height.load(SeqCst))?;
        window.set_position(
            WindowPos::Positioned(self.x.load(SeqCst)),
            WindowPos::Positioned(self.y.load(SeqCst)),
        );

        Ok(())
    }
}

pub struct WindowSystemIntegration {
    // sdl
    _video_subsystem: Mutex<sdl3::VideoSubsystem>,
    window: Mutex<sdl3::video::Window>,
    clipboard_util: ClipboardUtil,

    cursor: Mutex<Option<Cursor>>,

    pre_fullscreen_rect: CellRect,

    surface: Mutex<Option<Arc<Surface>>>,
}

unsafe impl Send for WindowSystemIntegration {}
unsafe impl Sync for WindowSystemIntegration {}

impl WindowSystemIntegration {
    pub(crate) fn new<'a>(
        create_info: &WindowCreateInfo,
        context: &Sdl,
    ) -> Result<Arc<WindowSystemIntegration>> {
        // create video subsystem
        let video_subsystem = context.video().map_err(|s| anyhow::Error::msg(s))?;

        // query display count
        let display_count = match video_subsystem.displays() {
            Ok(displays) => displays.len(),
            Err(_) => 0,
        };

        if display_count == 0 {
            return Err(anyhow::Error::msg("No display where detected!"));
        }

        // // query display information
        // let mut displays = Vec::with_capacity(display_count as usize);

        // for display in video_subsystem
        //     .displays()
        //     .map_err(|_| anyhow!("failed querying displays"))?
        // {
        //     let rect = display.get_bounds().map_err(|s| anyhow::Error::msg(s))?;
        //     let name = display.get_name().map_err(|s| anyhow::Error::msg(s))?;

        //     let display = Display {
        //         name,

        //         x: rect.x(),
        //         y: rect.y(),
        //         w: rect.width(),
        //         h: rect.height(),
        //     };

        //     displays.push(display);
        // }

        // build window
        let mut window_builder =
            video_subsystem.window(&create_info.title, create_info.width, create_info.height);
        window_builder.resizable().vulkan();
        window_builder.position_centered();
        let window = window_builder.build()?;

        let rect = CellRect::default();
        rect.update_from_window(&window);

        let wsi = WindowSystemIntegration {
            clipboard_util: video_subsystem.clipboard(),
            _video_subsystem: Mutex::new(video_subsystem),
            window: Mutex::new(window),

            cursor: Mutex::new(None),

            pre_fullscreen_rect: rect,

            surface: Mutex::new(None),
        };

        if create_info.fullscreen {
            wsi.set_fullscreen(true)?;
        }

        Ok(Arc::new(wsi))
    }

    pub fn window_size(&self) -> (u32, u32) {
        self.window.lock().unwrap().size()
    }

    pub fn is_fullscreen(&self) -> bool {
        match self.window.lock().unwrap().fullscreen_state() {
            FullscreenType::Desktop => false,
            FullscreenType::True => true,
            FullscreenType::Off => false,
        }
    }

    pub fn set_fullscreen(&self, fullscreen: bool) -> Result<()> {
        let mut window = self.window.lock().unwrap();

        if fullscreen {
            // store window information
            self.pre_fullscreen_rect.update_from_window(&window);

            // set fullscreen size to fit display
            let display = window.get_display().map_err(|s| anyhow!(s))?;
            let rect = display.get_bounds().map_err(|s| anyhow!(s))?;

            set_window_size(&mut window, rect.w as u32, rect.h as u32)?;

            window.set_bordered(false);

            // change fullscreen mode
            window
                .set_fullscreen(true)
                .map_err(|s| anyhow::Error::msg(s))?;
        } else {
            // change fullscreen mode
            window
                .set_fullscreen(false)
                .map_err(|s| anyhow::Error::msg(s))?;

            // force window borders
            window.set_bordered(true);

            // update window values
            self.pre_fullscreen_rect.update_to_window(&mut window)?;
        }

        Ok(())
    }

    pub fn show_simple_info_box(&self, title: &str, message: &str) -> Result<()> {
        self.show_simple_message_box(MessageBoxFlag::INFORMATION, title, message)
    }

    pub fn show_simple_warning_box(&self, title: &str, message: &str) -> Result<()> {
        self.show_simple_message_box(MessageBoxFlag::WARNING, title, message)
    }

    pub fn show_simple_error_box(&self, title: &str, message: &str) -> Result<()> {
        self.show_simple_message_box(MessageBoxFlag::ERROR, title, message)
    }

    pub fn clipboard_content(&self) -> Result<Option<String>> {
        Ok(if self.clipboard_util.has_clipboard_text() {
            Some(
                self.clipboard_util
                    .clipboard_text()
                    .map_err(|err| Error::msg(err))?,
            )
        } else {
            None
        })
    }

    #[inline]
    fn show_simple_message_box(
        &self,
        flags: MessageBoxFlag,
        title: &str,
        message: &str,
    ) -> Result<()> {
        let window = self.window.lock().unwrap();

        show_simple_message_box(flags, title, message, Some(window.deref()))?;

        Ok(())
    }

    pub fn set_opacity(&self, opacity: f32) -> Result<()> {
        self.window
            .lock()
            .unwrap()
            .set_opacity(opacity)
            .map_err(|s| anyhow::Error::msg(s))?;

        Ok(())
    }

    pub fn set_icon<T: AsRef<Path>>(&self, path: T) -> Result<()> {
        // let texture = match image::open(path) {
        //     Ok(tex) => tex.to_rgba(),
        //     Err(err) => create_error!(format!("error loading image: {}", err)),
        // };

        // let (width, height) = texture.dimensions();
        // let mut texture_data = texture.into_raw();

        // let surface = SDL_Surface::from_data(
        //     &mut texture_data,
        //     width,
        //     height,
        //     width * 4,
        //     PixelFormatEnum::RGBA8888,
        // )?;

        let surface = SDL_Surface::load_bmp(path).map_err(|s| anyhow::Error::msg(s))?;

        self.window.lock().unwrap().set_icon(surface);

        Ok(())
    }

    pub fn set_cursor<T: AsRef<Path>>(&self, path: T) -> Result<()> {
        // let texture = match image::open(path) {
        //     Ok(tex) => tex.to_rgba(),
        //     Err(err) => create_error!(format!("error loading image: {}", err)),
        // };

        // let (width, height) = texture.dimensions();
        // let mut texture_data = texture.into_raw();

        // let surface = SDL_Surface::from_data(
        //     &mut texture_data,
        //     width,
        //     height,
        //     width * 4,
        //     PixelFormatEnum::RGBA8888,
        // )?;

        let surface = SDL_Surface::load_bmp(path).map_err(|s| anyhow::Error::msg(s))?;
        let cursor = Cursor::from_surface(surface, 0, 0).map_err(|s| anyhow::Error::msg(s))?;

        cursor.set();

        *self.cursor.lock().unwrap() = Some(cursor);

        Ok(())
    }

    pub fn create_vulkan_surface(&self, instance: &Arc<Instance>) -> Result<()> {
        let vk_surface = unsafe {
            core::mem::transmute(
                self.window
                    .lock()
                    .unwrap()
                    .vulkan_create_surface(core::mem::transmute(instance.vk_handle().raw()))
                    .map_err(|s| anyhow::Error::msg(s))?,
            )
        };

        *self.surface.lock().unwrap() = Some(Surface::from_vk_surface(vk_surface, instance));

        Ok(())
    }

    pub fn surface(&self) -> Arc<Surface> {
        self.surface.lock().unwrap().as_ref().unwrap().clone()
    }

    pub(crate) fn activate_vulkan_instance_extensions(
        &self,
        extensions: &mut InstanceExtensions,
    ) -> Result<()> {
        let driver = self._video_subsystem.lock().unwrap().current_video_driver();

        match driver {
            "android" => extensions.android_surface = true,
            "cocoa" => extensions.macos_surface = true,
            "x11" => extensions.xlib_surface = true,
            "wayland" => extensions.wayland_surface = true,
            "windows" => extensions.win32_surface = true,

            _ => panic!("unsupported video driver: {driver}"),
        }

        extensions.surface = true;

        Ok(())
    }

    pub(crate) fn activate_vulkan_device_extensions(
        &self,
        extensions: &mut DeviceExtensions,
    ) -> Result<()> {
        extensions.swapchain = true;

        Ok(())
    }
}

impl std::fmt::Debug for WindowSystemIntegration {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "WindowSystemIntegration {{ }}")
    }
}

/// helper function to wrap SDL2 error types
#[inline]
fn set_window_size(window: &mut sdl3::video::Window, width: u32, height: u32) -> Result<()> {
    window.set_size(width, height)?;

    Ok(())
}
